
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <sys/times.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkvaria.h"
#include "pkscanner.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"
#include "multibs.h"
#include "bsmesh.h"
#include "g2blendingd.h"
#include "egholed.h"
#include "mengerc.h"
#include "bsfile.h"
#include "pkrender.h"
#include "xgedit.h"
#include "xgledit.h"

#include "widgets.h"
#include "editor.h"
#include "editor_bezc.h"
#include "editor_bsc.h"
#include "editor_bezp.h"
#include "editor_bsp.h"
#include "editor_bsm.h"
#include "editor_bsh.h"
#include "pozwalaj.h"
#include "edlight.h"


#define MY_ID_SPACE      6
#define MAXSTRINGLGT   127

#define cmd_0            0
#define cmd_ACTIVATE     0
#define cmd_ACTIVATING   1
#define cmd_ALL          2
#define cmd_CAMERA       3
#define cmd_DEACTIVATE   4
#define cmd_DELETE       5
#define cmd_EXIT         6
#define cmd_FRAME        7
#define cmd_INTSTR       8
#define cmd_LIGHT        9
#define cmd_LIST        10
#define cmd_OPENFILE    11
#define cmd_PURGE       12
#define cmd_RENDER      13
#define cmd_RESIZE      14
#define cmd_ROTCV       15
#define cmd_ROTCX       16
#define cmd_ROTCY       17
#define cmd_ROTCZ       18
#define cmd_ROTGV       19
#define cmd_ROTGX       20
#define cmd_ROTGY       21
#define cmd_ROTGZ       22
#define cmd_RUN         23
#define cmd_SETF        24
#define cmd_SHADOWS     25
#define cmd_VIEWPORT    26
#define cmd_WRITETIFF   27
#define cmd_ZOOM        28
#define NKEYWORDS       29

/* must be ASCII-alphabetical */
static const char *cmd_name[NKEYWORDS] =
  {"activate",
   "activating",
   "all",
   "camera",
   "deactivate",
   "delete",
   "exit",
   "frame",
   "intstr",
   "light",
   "list",
   "openfile",
   "purge",
   "render",
   "resize",
   "rotVC",
   "rotXC",
   "rotYC",
   "rotZC",
   "rotVG",
   "rotXG",
   "rotYG",
   "rotZG",
   "run",
   "setF",
   "shadows",
   "viewport",
   "writetiff",
   "zoom"};


static pks_batchexec _batch;
boolean              command_pending = false;

/* ////////////////////////////////////////////////////////////////////////// */
boolean myAppIdentFind ( pks_scanner *sc )
{
  int r;

  if ( (r = pks_BinSearchNameTable ( NKEYWORDS, 0, cmd_name,
                                     sc->nextname )) != PKS_SYMB_ERROR ) {
    sc->nextinteger = r;
    sc->id_space = MY_ID_SPACE;
    return true;
  }
  else
    return false;
} /*myAppIdentFind*/

boolean myAppIdentCalc ( pks_scanner *sc, boolean ignore, pks_calcval *x )
{
  if ( sc->nextsymbol == PKS_SYMB_IDENT && sc->id_space == MY_ID_SPACE ) {
    if ( sc->nextinteger == cmd_FRAME ) {
      if ( !ignore ) {
        x->valtype = PKS_SYMB_INTEGER;
        x->intval = pictcnt;
        x->floatval = (double)pictcnt;
      }
      sc->GetNextSymbol ( sc );  /* skip the variable name */
      return true;
    }
  }
  return false;
} /*myAppIdentCalc*/

static boolean myReadVector3d ( pks_scanner *sc, vector3d *v )
{
  pks_calcval x;

  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    return false;
  sc->GetNextSymbol ( sc );
  if ( !pkscalc_Expression ( sc, &x ) )
    return false;
  v->x = x.floatval;
  if ( sc->nextsymbol != PKS_SYMB_COMMA )
    return false;
  sc->GetNextSymbol ( sc );
  if ( !pkscalc_Expression ( sc, &x ) )
    return false;
  v->y = x.floatval;
  if ( sc->nextsymbol != PKS_SYMB_COMMA )
    return false;
  sc->GetNextSymbol ( sc );
  if ( !pkscalc_Expression ( sc, &x ) )
    return false;
  v->z = x.floatval;
  if ( sc->nextsymbol != PKS_SYMB_RBRACE )
    return false;
  sc->GetNextSymbol ( sc );
  return true;
} /*myReadVector3d*/

static boolean myCameraCommand ( pks_scanner *sc )
{
  int         id;
  pks_calcval x;
  short       width, height;
  vector3d    v;
  CameraRecd  *CPos;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol == PKS_SYMB_IDENT && sc->id_space == MY_ID_SPACE ) {
    id = sc->nextinteger;
    CPos = &g00win3D.CPos[3];
    switch ( id ) {
  case cmd_VIEWPORT:
      sc->GetNextSymbol ( sc );
      if ( sc->nextsymbol != PKS_SYMB_INTEGER ) return false;
      width = (short)sc->nextinteger;
      sc->GetNextSymbol ( sc );
      if ( sc->nextsymbol != PKS_SYMB_INTEGER ) return false;
      height = (short)sc->nextinteger;
      sc->GetNextSymbol ( sc );
      if ( geom00win == geom00win3D )
        xge_FourSetQuarterSize ( &g00win3D.fww, 3, width, height );
      else
        return false;
      break;
  case cmd_ROTCV:
  case cmd_ROTGV:
      sc->GetNextSymbol ( sc );
      if ( !pkscalc_Expression ( sc, &x ) ) return false;
      if ( !myReadVector3d ( sc, &v ) ) return false;
      NormalizeVector3d ( &v );
      if ( id == cmd_ROTCV )
        CameraRotVCd ( CPos, &v, x.floatval );
      else
        CameraRotVGd ( CPos, &v, x.floatval );
      break;
  case cmd_ROTCX:
  case cmd_ROTCY:
  case cmd_ROTCZ:
  case cmd_ROTGX:
  case cmd_ROTGY:
  case cmd_ROTGZ:
  case cmd_SETF:
  case cmd_ZOOM:
      sc->GetNextSymbol ( sc );
      if ( !pkscalc_Expression ( sc, &x ) ) return false;
      switch ( id ) {
    case cmd_ROTCX: CameraRotXCd ( CPos, x.floatval );  break;
    case cmd_ROTCY: CameraRotYCd ( CPos, x.floatval );  break;
    case cmd_ROTCZ: CameraRotZCd ( CPos, x.floatval );  break;
    case cmd_ROTGX: CameraRotXGd ( CPos, x.floatval );  break;
    case cmd_ROTGY: CameraRotYGd ( CPos, x.floatval );  break;
    case cmd_ROTGZ: CameraRotZGd ( CPos, x.floatval );  break;
    case cmd_SETF:  CameraSetFd ( CPos, x.floatval );   break;
    case cmd_ZOOM:  CameraZoomd ( CPos, x.floatval );   break;
    default:        return false;
      }
      break;
  default:
      return false;
    }
    SetupCameraParamWidgets ( false );
    return true;
  }
  else
    return false;
} /*myCameraCommand*/

static boolean myLightCommand ( pks_scanner *sc )
{
  int         id, l;
  pks_calcval x;
  vector3d    v;
  trans3d     tr;
  CameraRecd  *CPos;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_INTEGER ) return false;
  l = sc->nextinteger;
  if ( l < 0 || l >= RENDERING_NLIGHTS )
    return false;
  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol == PKS_SYMB_IDENT && sc->id_space == MY_ID_SPACE ) {
    id = sc->nextinteger;
    sc->GetNextSymbol ( sc );
    IdentTrans3d ( &tr );
    CPos = &g00win3D.CPos[3];
    switch ( id ) {
  case cmd_ROTCV:
  case cmd_ROTGV:
      if ( !pkscalc_Expression ( sc, &x ) ) return false;
      if ( !myReadVector3d ( sc, &v ) ) return false;
      NormalizeVector3d ( &v );
      if ( id == cmd_ROTCV ) {
        EulerRotTrans3d ( &tr, -CPos->psi, -CPos->theta, -CPos->phi );
        RotVTrans3d ( &tr, &v, x.floatval );
        EulerRotTrans3d ( &tr, CPos->phi, CPos->theta, CPos->psi );
      }
      else
        RotVTrans3d ( &tr, &v, x.floatval );
      goto rotate_light_dir;
  default:
      if ( !pkscalc_Expression ( sc, &x ) ) return false;
      switch ( id ) {
    case cmd_ROTCX:
        EulerRotTrans3d ( &tr, -CPos->psi, -CPos->theta, -CPos->phi );
        RotXTrans3d ( &tr, x.floatval );
        EulerRotTrans3d ( &tr, CPos->phi, CPos->theta, CPos->psi );
        goto rotate_light_dir;
    case cmd_ROTCY:
        EulerRotTrans3d ( &tr, -CPos->psi, -CPos->theta, -CPos->phi );
        RotYTrans3d ( &tr, x.floatval );
        EulerRotTrans3d ( &tr, CPos->phi, CPos->theta, CPos->psi );
        goto rotate_light_dir;
    case cmd_ROTCZ:
        EulerRotTrans3d ( &tr, -CPos->psi, -CPos->theta, -CPos->phi );
        RotZTrans3d ( &tr, x.floatval );
        EulerRotTrans3d ( &tr, CPos->phi, CPos->theta, CPos->psi );
        goto rotate_light_dir;
    case cmd_ROTGX:
        RotXTrans3d ( &tr, x.floatval );
        goto rotate_light_dir;
    case cmd_ROTGY:
        RotYTrans3d ( &tr, x.floatval );
        goto rotate_light_dir;
    case cmd_ROTGZ:
        RotZTrans3d ( &tr, x.floatval );
        goto rotate_light_dir;
    default:
        return false;
      }
rotate_light_dir:
      TransVector3d ( &tr, &render_light_dir[l], &render_light_dir[l] );
      NormalizeVector3d ( &render_light_dir[l] );
    }
  }
  else {
    if ( !pkscalc_Expression ( sc, &x ) )
      return false;
    if ( !myReadVector3d ( sc, &v ) )
      return false;
    if ( x.floatval < 0.0 )
      x.floatval = 0.0;
    else if ( x.floatval > 1.0 )
      x.floatval = 1.0;
    NormalizeVector3d ( &v );
    render_light_dir[l] = v;
    render_light_int[l] = x.floatval;
  }
  rendered_picture = false;
  return true;
} /*myLightCommand*/

static boolean mySetSwOnOff ( pks_scanner *sc, int key )
{
  pks_calcval x;
  boolean     sw;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_EQ )
    return false;
  sc->GetNextSymbol ( sc );
  if ( !pkscalc_Expression ( sc, &x ) )
    return false;
  sw = x.valtype == PKS_SYMB_INTEGER ? x.intval != 0 : x.floatval != 0.0;
  switch ( key ) {
case cmd_ACTIVATING:
    activate_readin_objects = sw;
    break;
case cmd_SHADOWS:
    swShadows = sw;
    break;
default:
    break;
  }
  return true;
} /*mySetSwOnOff*/

static boolean GetNextString ( pks_scanner *sc, char *name, int maxlgt )
{
  char format[64];

  switch ( sc->nextsymbol ) {
case PKS_SYMB_STRING:
    strncpy ( name, sc->nextname, maxlgt );
    name[maxlgt-1] = 0;
    sc->GetNextSymbol ( sc );
    return true;
case PKS_SYMB_IDENT:
    if ( sc->id_space != MY_ID_SPACE || sc->nextinteger != cmd_INTSTR )
      return false;
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_STRING )
      return false;
    strncpy ( format, sc->nextname, 20 );
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_INTEGER )
      return false;
    sprintf ( name, format, sc->nextinteger );
    sc->GetNextSymbol ( sc );
    return true;
default:
    return false;
  }
} /*GetNextString*/

static boolean myActivate ( pks_scanner *sc, boolean key )
{
  char name[MAXSTRINGLGT+1];

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol == PKS_SYMB_IDENT ) {
    if ( sc->id_space == MY_ID_SPACE && sc->nextinteger == cmd_ALL ) {
      sc->GetNextSymbol ( sc );
      if ( GeomObjectActivateAll ( key ) )
        rendered_picture = false;
      return true;
    }
  }
  if ( GetNextString ( sc, name, MAXSTRINGLGT ) ) {
    if ( GeomObjectActivateByName ( name, key ) )
      rendered_picture = false;
    return true;
  }
  else
    return false;
} /*myActivate*/

static boolean myDelete ( pks_scanner *sc )
{
  char name[MAXSTRINGLGT+1];

  sc->GetNextSymbol ( sc );
  if ( GetNextString ( sc, name, MAXSTRINGLGT ) ) {
    if ( GeomObjectDeleteByName ( name ) ) {
      UpdateObjectNameList ();
      SetupObjectSpecificMenus ( current_go );
      rendered_picture = false;
    }
    return true;
  }
  else
    return false;
} /*myDelete*/

static boolean myFrameCommand ( pks_scanner *sc )
{
  pks_calcval x;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_EQ )
    return false;
  sc->GetNextSymbol ( sc );
  if ( pkscalc_Expression ( sc, &x ) ) {
    if ( x.valtype == PKS_SYMB_INTEGER ) {
      pictcnt = min ( x.intval, 9999 );
      pictcnt = max ( 0, pictcnt );
      return true;
    }
  }
  return false;
} /*myFrameCommand*/

static boolean myOpenFile ( pks_scanner *sc )
{
  char name[MAXSTRINGLGT+1];

  sc->GetNextSymbol ( sc );
  if ( GetNextString ( sc, name, MAXSTRINGLGT ) ) {
    if ( GeomObjectReadFile ( name, Popup01CameraReader ) ) {
      current_go = first_go;
      rendered_picture = false;
      return true;
    }
    else
      return false;   
  }
  else
    return false;
} /*myOpenFile*/

static boolean myPurge ( pks_scanner *sc )
{
  PurgeTheObjectList ();
  sc->GetNextSymbol ( sc );
  rendered_picture = false;
  return true;
} /*myPurge*/

static boolean myRunBatchFile ( pks_scanner *sc, boolean from_batch )
{
  char fn[MAXSTRINGLGT+1];
  char *params;

  if ( from_batch || command_pending )
    return false;
  sc->GetNextSymbol ( sc );  /* skip "run", get batch file name */
  if ( sc->nextsymbol != PKS_SYMB_STRING )
    return false;
  pkv_CopyString ( fn, sc->nextname, MAXSTRINGLGT );
  params = &sc->infile->inbuffer[sc->infile->inbufpos];
  if ( pksbatch_InitExecution ( &_batch, PKS_OPT_C_OPERATORS,
                                fn, 0, MAXSTRINGLGT, "//", "\n",
                                params, NULL, myAppIdentFind, myAppIdentCalc,
                                NULL ) ) {
    command_pending = true;
    xge_PostIdleCommand ( IDLE_COMMAND_NEXT_BATCH_COMMAND, 0, 0 );
    return true;
  }
  else
    return false;
} /*myRunBatchFile*/

static boolean myListBatchFile ( pks_scanner *sc, boolean from_batch )
{
  char fn[MAXSTRINGLGT+5];
  char *params;
  FILE *f;

  if ( from_batch || command_pending )
    return false;
  sc->GetNextSymbol ( sc );  /* skip "list", get batch file name */
  if ( sc->nextsymbol != PKS_SYMB_STRING )
    return false;
  pkv_CopyString ( fn, sc->nextname, MAXSTRINGLGT );
  params = &sc->infile->inbuffer[sc->infile->inbufpos];
  if ( pksbatch_InitExecution ( &_batch, PKS_OPT_C_OPERATORS,
                                fn, 0, MAXSTRINGLGT, "//", "\n",
                                params, NULL, myAppIdentFind, myAppIdentCalc,
                                NULL ) ) {
    strcpy ( &fn[strlen(fn)], ".lst" );
    if ( (f = fopen ( fn, "w+" ) ) ) {
      pksbatch_ListCompiledCode ( &_batch, f );
      fclose ( f );
      pksbatch_EndExecution ( &_batch );
      return true;
    }
  }
  return false;
} /*myListBatchFile*/

static boolean myStartRendering ( pks_scanner *sc )
{
  sc->GetNextSymbol ( sc );
  StartRendering ();
  while ( sc->nextsymbol == PKS_SYMB_SEMICOLON )
    sc->GetNextSymbol ( sc );
  return true;
} /*myStartRendering*/

boolean BreakBatchExecution ( void )
{
  char s[20];

  if ( command_pending ) {
    SetStatusText ( win1, "", true );
    pksbatch_EndExecution ( &_batch );
    command_pending = false;
    sprintf ( s, "time = %3.1f", pkv_Seconds ( _batch.ticks ) );
    return true;
  }
  else {
    DisplayErrorMessage1 ( ErrorMsgNoCommandToBreak, -1 );
    return false;
  }
} /*BreakBatchExecution*/

static boolean myWriteTiff ( pks_scanner *sc )
{
  char s[MAXSTRINGLGT+1];

  sc->GetNextSymbol ( sc );
  sprintf ( s, "%s %d", cmd_name[cmd_WRITETIFF], pictcnt );
  SetStatusText ( win1, s, true );
  WriteTiffFile ();
  return true;
} /*myWriteTiff*/

static boolean myResize ( pks_scanner *sc )
{
  int win, ww, wh;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_INTEGER )
    return false;
  win = sc->nextinteger;
  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_INTEGER )
    return false;
  ww = sc->nextinteger;
  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_INTEGER )
    return false;
  wh = sc->nextinteger;
  sc->GetNextSymbol ( sc );
  if ( ww >= xge_WIDTH && ww <= xge_MAX_WIDTH &&
       wh >= xge_HEIGHT && wh <= xge_MAX_HEIGHT ) {
    if ( win == win0 || win == win1 ) {
      xge_SetWindow ( win );
      xgeResizeWindow ( ww, wh );
      if ( win == win0 )
        ResizeWindow0 ( ww, wh );
      else
        ResizeWindow1 ( ww, wh );
    }
    /* else do nothing */
  }
  return true;
} /*myResize*/

static boolean myCommand ( pks_scanner *sc,
                           boolean from_batch, boolean *post_idle )
{
  boolean adv;

  adv = false;
  *post_idle = true;
  for (;;) {
    switch ( sc->nextsymbol ) {
  case PKS_SYMB_IDENT:
      SetStatusText ( win1, sc->nextname, true );
      switch ( sc->id_space ) {
    case MY_ID_SPACE:
        switch ( sc->nextinteger ) {
      case cmd_ACTIVATE:
          if ( myActivate ( sc, true ) )
            break;
          else
            goto failure;
      case cmd_ACTIVATING:
          if ( mySetSwOnOff ( sc, cmd_ACTIVATING ) )
            break;
          else
            goto failure;
      case cmd_CAMERA:
          if ( myCameraCommand ( sc ) )
            break;
          else
            goto failure;
      case cmd_DEACTIVATE:
          if ( myActivate ( sc, false ) )
            break;
          else
            goto failure;
      case cmd_DELETE:
          sc->GetNextSymbol ( sc );
          if ( myDelete ( sc ) )
            break;
          else
            goto failure;
      case cmd_EXIT:
          BreakBatchExecution ();
          *post_idle = false;
          return true;
      case cmd_FRAME:
          if ( myFrameCommand ( sc ) )
            break;
          else
            goto failure;
      case cmd_LIGHT:
          if ( myLightCommand ( sc ) )
            break;
          else
            goto failure;
      case cmd_LIST:
          return myListBatchFile ( sc, from_batch );
      case cmd_OPENFILE:
          if ( myOpenFile ( sc ) )
            break;
          else
            goto failure;
      case cmd_PURGE:
          if ( myPurge ( sc ) )
            break;
          else
            goto failure;
      case cmd_RENDER:
          if ( myStartRendering ( sc ) ) {
            *post_idle = false;
            return true;
          }
          else
            goto failure;
      case cmd_RESIZE:
          if ( myResize ( sc ) )
            break;
          else
            goto failure;
      case cmd_RUN:
          return myRunBatchFile ( sc, from_batch );
      case cmd_SHADOWS:
          if ( mySetSwOnOff ( sc, cmd_SHADOWS ) )
            break;
          else
            goto failure;
      case cmd_WRITETIFF:
          if ( myWriteTiff ( sc ) )
            break;
          else
            goto failure;
      default:
          goto failure;
        }
        break;

    default:
        goto failure;
      }
      break;

  case PKS_SYMB_COMMA:
      sc->GetNextSymbol ( sc );
      continue;

  case PKS_SYMB_SEMICOLON:
      sc->GetNextSymbol ( sc );
      return true;

  case PKS_SYMB_EOF:
      if ( from_batch )
        return BreakBatchExecution ();
      else
        return true;

  case PKS_SYMB_ERROR:
      goto failure;

  default:
      if ( adv )
        return true;
      else
        goto failure;
    }
    adv = true;
  }

failure:
printf ( "****\n" );
  if ( from_batch )
    BreakBatchExecution ();
  return false;
} /*myCommand*/

boolean ExecNextBatchCommand ( void )
{
  boolean post_idle;

  if ( !command_pending )
    return false;
  if ( myCommand ( &_batch.csc, true, &post_idle ) ) {
    if ( post_idle )
      xge_PostIdleCommand ( IDLE_COMMAND_NEXT_BATCH_COMMAND, 0, 0 );
    return true;
  }
  else
    return false;
} /*ExecNextBatchCommand*/

boolean ExecCommand ( pks_scanner *sc, char *command )
{
  boolean post_idle;

  return myCommand ( sc, false, &post_idle );
} /*ExecCommand*/

