
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <limits.h>
#include <string.h>

#include "pkvaria.h"
#include "pknum.h"  
#include "pkgeom.h" 
#include "multibs.h"


static boolean _mbs_AtOneSided ( short int dim, short int deg,
                                 const double *points,
                                 double x0, double x1, double y0, double y1 )
{
  int i, j;

  if ( dim == 2 ) {
    for ( i = j = 0;  i <= deg;  i++, j += 2 )
      if ( points[j] > x0 )
        goto not_x0;
    return true;
not_x0:
    for ( i = j = 0;  i <= deg;  i++, j += 2 )
      if ( points[j] < x1 )
        goto not_x1;
    return true;
not_x1:
    for ( i = 0, j = 1;  i <= deg;  i++, j += 2 )
      if ( points[j] > y0 )
        goto not_y0;
    return true;
not_y0:
    for ( i = 0, j = 1;  i <= deg;  i++, j += 2 )
      if ( points[j] < y1 )
        return false;
  }
  else {
    for ( i = j = 0;  i <= deg;  i++, j += 3 )
      if ( points[j]/points[j+2] > x0 )
        goto not_X0;
    return true;
not_X0:
    for ( i = j = 0;  i <= deg;  i++, j += 3 )
      if ( points[j]/points[j+2] < x1 )
        goto not_X1;
    return true;
not_X1:
    for ( i = 0, j = 1;  i <= deg;  i++, j += 3 )
      if ( points[j]/points[j+1] > y0 )
        goto not_Y0;
    return true;
not_Y0:
    for ( i = 0, j = 1;  i <= deg;  i++, j += 3 )
      if ( points[j]/points[j+1] < y1 )
        return false;
  }
  return true;
} /*_mbs_AtOneSided*/

char mbs_ConstantRectangleIndexd ( int nelem, mbs_polycurved *poly,
                                   double x0, double x1, double y0, double y1 )
{
#define CONVERTP(dim,p,q) \
  { if ( dim == 2 ) memcpy ( q, p, sizeof(point2d) ); \
    else Point3to2d ( (point3d*)p, q ); }

  int     dim, deg, lkn, ku;
  int     i, j;
  double  *points, *bsbuf;
  point2d segment[3];
  boolean first;
  char    result;

  result = 1;
  first = true;
  for ( i = 0; i < nelem; i++ ) {
    dim = poly[i].cpdimen;
    deg = poly[i].degree;
    lkn = poly[i].lastknot;
    points = poly[i].points;
    if ( poly[i].knots ) { /* a B-spline curve */
      ku = mbs_NumKnotIntervalsd ( deg, lkn, poly[i].knots );
      bsbuf = pkv_GetScratchMemd ( dim*(deg+1)*ku );
      if ( !bsbuf )
        return -1;
      if ( !mbs_multiBSCurvesToBezd ( dim, 1, deg, lkn, poly[i].knots,
                                      0, points, &ku, NULL, NULL, 0, bsbuf ) ) {
        result = -1;
        goto free_bsbuf;
      }
      CONVERTP ( dim, bsbuf, &segment[1] )
      if ( first ) segment[2] = segment[1];
      else {
        if ( !_mbs_AtOneSided ( 2, 1, &segment[0].x, x0, x1, y0, y1 ) ) {
          result = 0;
          goto free_bsbuf;
        }
      }
      for ( j = 0; j < ku; j++ ) {
        if ( !_mbs_AtOneSided ( dim, deg, &bsbuf[j*dim*(deg+1)],
                                x0, x1, y0, y1 ) ) {
          result = 0;
          goto free_bsbuf;
        }
      }
      CONVERTP ( dim, &bsbuf[ku*dim*(deg+1)-dim], segment )
      pkv_SetScratchMemTop ( bsbuf );
    }
    else if ( lkn == -1 ) { /* a Bezier curve */
      CONVERTP ( dim, points, &segment[1] )
      if ( first ) segment[2] = segment[1];
      else if ( !_mbs_AtOneSided ( 2, 1, &segment[0].x, x0, x1, y0, y1 ) )
        return 0;
      if ( !_mbs_AtOneSided ( dim, deg, points, x0, x1, y0, y1 ) )
        return 0;
      CONVERTP ( dim, &points[dim*deg], segment );
    }
    else { /* a polyline made of lkn line segments */
      CONVERTP ( dim, points, &segment[1] )
      if ( first ) segment[2] = segment[1];
      else if ( !_mbs_AtOneSided ( 2, 1, &segment[0].x, x0, x1, y0, y1 ) )
        return 0;
      for ( j = 0; j < lkn; j++ )
        if ( !_mbs_AtOneSided ( dim, 1, &points[dim*j], x0, x1, y0, y1 ) )
          return 0;
      CONVERTP ( dim, &points[dim*lkn], segment );
    }
    if ( (first = (poly[i].closing != 0)) ) {
      segment[1] = segment[2];
      if ( !_mbs_AtOneSided ( 2, 1, &segment[0].x, x0, x1, y0, y1 ) )
        return 0;
    }
  }
  return 1;

free_bsbuf:
  pkv_SetScratchMemTop ( bsbuf );
  return result;
#undef CONVERTP
} /*mbs_ConstantRectangleIndexd*/

