
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <sys/times.h>

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

#include "pksprivate.h"
#include "pkbprivate.h"

/* the keywords must match the #defined symbols in pkscanner.h */
const char *pksbatch_cmdnames[pksbatch_NUMNAMES] =
  { "break",
    "case",
    "continue",
    "default",
    "do",
    "else",
    "for",
    "goto",
    "if",
    "label",
    "switch",
    "while" };

#define LAST_SEMICOLON ((auxcomp*)(batch->aux))->last_semicolon

static boolean _pksb_CompileKeyStatement ( pks_batchexec *batch,
                                           block_labels *blab,
                                           boolean inloop, boolean inswitch );

static boolean _pksb_CompileTokenSequence ( pks_batchexec *batch,
                                            block_labels *blab,
                                            short int *stop, short int *banned,
                                            boolean inloop, boolean inswitch );

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _pksb_AppIdentFind ( pks_scanner *sc )
{
  pks_batchexec *batch;
  int r;

  batch = (pks_batchexec*)sc->libdata;
  if ( (r = pks_BinSearchNameTable ( pkscalc_NUMNAMES, 0, pkscalc_fnames, 
                                     sc->nextname )) != PKS_SYMB_ERROR ) {
    sc->nextinteger = r;
    sc->id_space = ID_SPACE_CALC;
  }
  else if ( (r = pks_BinSearchNameTable ( pksbatch_NUMNAMES, 0, pksbatch_cmdnames,
                                          sc->nextname )) != PKS_SYMB_ERROR ) {
    sc->nextinteger = r;
    sc->id_space = ID_SPACE_BATCH;
  }
  else if ( batch->AppIdentFind ( sc ) )
    ;
  else {
    sc->id_space = ID_SPACE_NONE;
    sc->nextinteger = -1;
  }
  return true;
} /*_pksb_AppIdentFind*/

/* ////////////////////////////////////////////////////////////////////////// */
/* Macro parameters may be single tokens or sequences of tokens enclosed in   */
/* braces, e.g. { token1 token2 ... }                                         */
/* all open braces must be closed and the outer braces are removed.           */
/* No preprocessing is done to the parameters, no evaluation of constant      */
/* expression etc., and no preprocessor command like #define, #if etc. are    */
/* allowed. The parameters are converted to proprocessor macros named ARG1,   */
/* ARG2 etc., and the macro ARGC is the integer number to denote the total    */
/* number of the parameters given to the batch file.                          */
static boolean _pksbatch_ProcessParameters ( pks_batchexec *batch,
                                             unsigned int options,
                                             char *params,
                                             int maxnamelength )
{
  void        *sp;
  int         argc;
  char        mcn[10];
  pks_macro   *mcr, *mcri;
  pks_token   end;
  char        *mbuf;
  int         brn, bufp;

  if ( !params )  /* do not bother */
    return true;
  if ( !pks_InitScanner ( &batch->csc, options, strlen(params), params,
                maxnamelength, NULL, NULL, NULL, NULL, NULL, NULL, NULL ) )
    return false;
  sp = pkv_GetScratchMemTop ();
  mbuf = pkv_GetScratchMem ( PKS_MACRO_BUF_LENGTH );
  if ( !mbuf )
    goto failure;
  mcri = (pks_macro*)mbuf;
  batch->csc.GetNextSymbol ( &batch->csc );
  for ( argc = 0; ; argc ++ ) {
    bufp = sizeof(pks_macro);
    memset ( mcri, 0, bufp+1 );
        /* generate the parameter macro name */
    sprintf ( mcn, "ARG%d", argc+1 );
    if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                           mcn, strlen(mcn)+1 ) ) {
      batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
      goto failure;
    }
    mcri->params = mcri->tokens = bufp;
    switch ( batch->csc.nextsymbol ) {
  case PKS_SYMB_EOF:
      goto finish;
  case PKS_SYMB_LBRACE:
      brn = 1;
      do {
        batch->csc.GetNextSymbol ( &batch->csc );
        switch ( batch->csc.nextsymbol ) {
    case PKS_SYMB_LBRACE:  brn ++;  break;
    case PKS_SYMB_RBRACE:  brn --;  break;
    case PKS_SYMB_EOF:     goto failure;
    default:               break;
        }
        if ( !_pks_AddTheMacroToken ( &batch->csc, mbuf, &bufp ) ) {
          batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
          goto failure;
        }
      } while ( brn > 0 );
      goto enter;
  case PKS_SYMB_RBRACE:
      goto failure;
  default:
      if ( !_pks_AddTheMacroToken ( &batch->csc, mbuf, &bufp ) ) {
        batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
        goto failure;
      }
enter:
      end.any.symbol = PKS_SYMB_NONE;
      end.any.length = sizeof(pks_token_any);
      if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                             (char*)&end, end.any.length ) ) {
        batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
        goto failure;
      }
      batch->csc.GetNextSymbol ( &batch->csc );
      mcr = malloc ( bufp+1 );
      if ( !mcr ) {
        batch->fsc.error_code = PRERROR_NO_MEMORY;
        goto failure;
      }
      memcpy ( mcr, mbuf, bufp );
      ((char*)mcr)[bufp] = 0;
      AVLInsert ( &batch->fsc.macros, (void*)mcr );
      break;
    }
  }

finish:
        /* enter the macro ARGC expanding to the integer number of parameters */
  bufp = sizeof(pks_macro);
  memset ( mcri, 0, bufp+1 );
  strcpy ( mcn, "ARGC" );
  if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                         mcn, strlen(mcn)+1 ) ) {
    batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
    goto failure;
  }
  mcri->params = mcri->tokens = bufp;
  end.integer.symbol = PKS_SYMB_INTEGER;
  end.integer.length = sizeof(pks_token_int);
  end.integer.intvalue = argc;
  if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                         (char*)&end, end.integer.length ) ) {
    batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
    goto failure;
  }
  end.any.symbol = PKS_SYMB_NONE;
  end.any.length = sizeof(pks_token_any);
  if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                         (char*)&end, end.any.length ) ) {
    batch->fsc.error_code = PRERROR_BUFFER_OVERFLOW;
    goto failure;
  }
  mcr = malloc ( bufp+1 );
  if ( !mcr ) {
    batch->fsc.error_code = PRERROR_NO_MEMORY;
    goto failure;
  }
  memcpy ( mcr, mbuf, bufp );
  ((char*)mcr)[bufp] = 0;
  AVLInsert ( &batch->fsc.macros, (void*)mcr );
        /* cleanup */
  pks_ShutDownScanner ( &batch->csc );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pks_ShutDownScanner ( &batch->csc );
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_pksbatch_ProcessParameters*/

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _pksb_IdentHook ( pks_scanner *sc, boolean ignore,
                                 pks_calcval *value )
{
  pks_batchexec *batch;

  value->valtype = PKS_SYMB_ERROR;
  if ( !(batch = (pks_batchexec*)sc->libdata) )
    return false;
  return batch->AppIdentCalc ( sc, ignore, value );
} /*_pksb_IdentHook*/

static boolean _pksb_AddToCode ( pks_batchexec *batch, void *data, int nbytes )
{
  return  pkv_AppendDataR ( &batch->code, &batch->buflgt, &batch->codelgt,
                            PKS_SCANNER_BUF_LENGTH, data, nbytes );
} /*_pksb_AddToCode*/

static boolean _pksb_AddSymbolToCode ( pks_batchexec *batch,
                                       void *data, int size )
{
  pks_scanner *sc;
  unsigned int place;

  sc = &batch->fsc;
  place = PKS_SCANNER_MAX_LINE_LENGTH*sc->nsline + sc->nscolumn;
  if ( !_pksb_AddToCode ( batch, &place, sizeof(unsigned int) ) )
    return false;
  return _pksb_AddToCode ( batch, data, size );
} /*_pksb_AddSymbolToCode*/

static boolean _pksb_AddLiteralToCode ( pks_batchexec *batch )
{
  pks_scanner *sc;

  sc = &batch->fsc;
  switch ( sc->nextsymbol ) {
case PKS_SYMB_INTEGER:
case PKS_SYMB_CHAR:
case PKS_SYMB_OTHER:
    if ( !_pksb_AddToCode ( batch, &sc->nextinteger, sizeof(int) ) )
      return false;
    break;
case PKS_SYMB_FLOAT:
    if ( !_pksb_AddToCode ( batch, &sc->nextfloat, sizeof(double) ) )
      return false;
    break;
case PKS_SYMB_IDENT:
    if ( !_pksb_AddToCode ( batch, &sc->id_space, sizeof(short int) ) )
      return false;
    if ( !_pksb_AddToCode ( batch, &sc->nextinteger, sizeof(int) ) )
      return false;
    if ( !_pksb_AddToCode ( batch, sc->nextname, strlen(sc->nextname)+1 ) )
      return false;
    break;
case PKS_SYMB_STRING:
    if ( !_pksb_AddToCode ( batch, sc->nextname, strlen(sc->nextname)+1 ) )
      return false;
    break;
default:
    break;
  }
  return true;
} /*_pksb_AddLiteralToCode*/

static boolean _pksb_EnterForCounter ( pks_batchexec *batch, boolean *forc )
{
  pks_scanner *sc;
  auxcomp     *aux;
  forcntname  *fcn;
  forcounter  forct;
  short int   level;

  sc = &batch->fsc;
  aux = batch->aux;
  for ( fcn = aux->firstcnt, level = aux->fornesting-1;
        fcn && level >= 0;
        fcn = fcn->next, level-- ) {
    if ( !strcmp ( sc->nextname, fcn->name ) ) {
      batch->codelgt -= sizeof(unsigned int) + sizeof(short int);
      forct.token = PKS_SYMB_BATCH_VAR;
      forct.level = level;
      *forc = true;
      return _pksb_AddSymbolToCode ( batch, &forct, sizeof(forct) );
    }
  }
  *forc = false;
  return true;
} /*_pksb_EnterForCounter*/

/* ////////////////////////////////////////////////////////////////////////// */
static labeldesc* _pksb_NewLabelDesc ( labeldesc **nextlbl, char *name )
{
  labeldesc *lbl;
  int       lgt;

  lgt = name ? strlen(name) : 0;
  lbl = malloc ( sizeof(labeldesc) + lgt );
  if ( lbl ) {
    lbl->next = *nextlbl;
    lbl->loc = lbl->caseval = -1;
    if ( name )
      strcpy ( lbl->name, name );
    else
      lbl->name[0] = 0;
    *nextlbl = lbl;
  }
  return lbl;
} /*_pksb_NewLabelDesc*/

static void _pksb_DeleteLabelList ( labeldesc *lbl )
{
  labeldesc *l;

  while ( (l = lbl) ) {
    lbl = l->next;
    free ( l );
  }
} /*_pksb_DeleteLabelList*/

static void _pksb_DeleteLabelLists ( block_labels *blab )
{
  _pksb_DeleteLabelList ( blab->gotos );
  _pksb_DeleteLabelList ( blab->labels );
  _pksb_DeleteLabelList ( blab->breaks );
  _pksb_DeleteLabelList ( blab->continues );
  _pksb_DeleteLabelList ( blab->cases );
  memset ( blab, 0, sizeof(block_labels) );
} /*_pksb_DeleteLabelLists*/

static void _pksb_MergeLabelLists ( labeldesc **lbl1, labeldesc *lbl2 )
{
  labeldesc *l;

  while ( (l = lbl2) ) {
    lbl2 = l->next;
    l->next = *lbl1;
    *lbl1 = l;
  }
} /*_pksb_MergeLabelLists*/

static void _pksb_ResolveBreaks ( pks_batchexec *batch,
                                  int dest, labeldesc *br )
{
  char      *code;
  labeldesc *jmp;

  code = batch->code;
  while ( (jmp = br) ) {
    br = jmp->next;
    *(int*)(&code[jmp->loc]) = dest;
    free ( jmp );
  }
} /*_pksb_ResolveBreaks*/

static void _pksb_ResolveGotos ( pks_batchexec *batch,
                                 labeldesc **lbl, labeldesc **gtl )
{
  char      *code;
  labeldesc **gt, *jmp, *lb;
  int       dest;

  code = batch->code;
  while ( *lbl ) {
    dest = (*lbl)->loc;
    gt = gtl;
    while ( (jmp = *gt) ) {
      if ( strcmp ( (*lbl)->name, jmp->name ) )
        gt = &jmp->next;
      else {
        *gt = jmp->next;
        *(int*)(&code[jmp->loc]) = dest;
        free ( jmp );
      }
    }
    lb = *lbl;
    *lbl = lb->next;
    free ( lb );
  }
} /*_pksb_ResolveGotos*/

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _pksb_CompileInnerStatement ( pks_batchexec *batch,
                                             block_labels *blab,
                                             boolean inloop, boolean inswitch )
{
  pks_scanner *sc;
  short int   stop[3], banned[4];

  sc = &batch->fsc;
  switch ( sc->nextsymbol ) {
case PKS_SYMB_ERROR:
case PKS_SYMB_EOF:
    goto failure;

case PKS_SYMB_LBRACE:
        /* inner statement is a sequence between matching braces */
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_RBRACE ) {  /* might be empty */
      stop[0] = PKS_SYMB_RBRACE;
      stop[1] = PKS_SYMB_NONE;
      banned[0] = PKS_SYMB_RPAREN;
      banned[1] = PKS_SYMB_RBRACKET;
      banned[2] = PKS_SYMB_NONE;
      if ( !_pksb_CompileTokenSequence ( batch, blab, stop, banned,
                                         inloop, inswitch ) )
        goto failure;
    }
    if ( sc->nextsymbol != PKS_SYMB_RBRACE )
      goto failure;
    LAST_SEMICOLON = true;  /* ??? */
    break;

case PKS_SYMB_IDENT:
    if ( sc->id_space != ID_SPACE_BATCH )
      goto othercase;
    if ( !_pksb_AddSymbolToCode ( batch, &sc->nextsymbol, sizeof(short int) ) )
      goto failure;
    if ( !_pksb_CompileKeyStatement ( batch, blab, inloop, inswitch ) )
      goto failure;
    LAST_SEMICOLON = true;  /* ??? */
    return true;

default:
othercase:
      /* inner statement is a sequence until a closing semicolon */
    stop[0] = PKS_SYMB_SEMICOLON;
    stop[1] = PKS_SYMB_NONE;
    banned[0] = PKS_SYMB_RPAREN;
    banned[1] = PKS_SYMB_RBRACKET;
    banned[2] = PKS_SYMB_RBRACE;
    banned[3] = PKS_SYMB_NONE;
    if ( !_pksb_CompileTokenSequence ( batch, blab, stop, banned,
                                       inloop, inswitch ) )
      goto failure;
    if ( LAST_SEMICOLON )  /* the semicolon has already been skipped */
      return true;
    if ( sc->nextsymbol != PKS_SYMB_SEMICOLON )
      goto failure;
    LAST_SEMICOLON = true;
    break;
  }
  sc->GetNextSymbol ( sc );
  return true;

failure:
  return false;
} /*_pksb_CompileInnerStatement*/

static boolean _pksb_CompileExpression ( pks_batchexec *batch,
                                         short int *stop )
{
  short int    banned[2];
  short int    op;
  block_labels blab;

  memset ( &blab, 0, sizeof(block_labels) );
  banned[0] = PKS_SYMB_BATCH_CMD;
  banned[1] = PKS_SYMB_NONE;
  if ( !_pksb_CompileTokenSequence ( batch, &blab, stop, banned,
                                     false, false ) )
    return false;
  op = PKS_SYMB_NONE;
  return _pksb_AddSymbolToCode ( batch, &op, sizeof(short int) );
} /*_pksb_CompileExpression*/

static boolean _pksb_CompileParentExpr ( pks_batchexec *batch )
{
  pks_scanner *sc;
  short int stop[2];

  sc = &batch->fsc;
  if ( sc->nextsymbol != PKS_SYMB_LPAREN )
    goto failure;
  sc->GetNextSymbol ( sc );
  stop[0] = PKS_SYMB_RPAREN;
  stop[1] = PKS_SYMB_NONE;
  if ( !_pksb_CompileExpression ( batch, stop ) )
    goto failure;
  if ( sc->nextsymbol != PKS_SYMB_RPAREN )
    goto failure;
  sc->GetNextSymbol ( sc );
  return true;

failure:
  return false;
} /*_pksb_CompileParentExpr*/

static boolean _pksb_CompileIfStatement ( pks_batchexec *batch,
                                          block_labels *_blab, boolean inloop )
{
  pks_scanner   *sc;
  block_labels  blab;
  ifstatement   ifst, *_ifst;
  jumpstatement jmp, *_jmp;
  int           ptr, jptr;

  sc = &batch->fsc;
  memset ( &blab, 0, sizeof(block_labels) );
  ifst.token = PKS_SYMB_BATCH_CMD;
  ifst.key = PKS_BATCHCMD_IF;
  ifst.ifptr = ifst.elseptr = 0;
  ptr = batch->codelgt;
  if ( !_pksb_AddSymbolToCode ( batch, &ifst, sizeof(ifstatement) ) )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip if keyword */
  if ( !_pksb_CompileParentExpr ( batch ) )
    goto failure;
  ifst.ifptr = batch->codelgt;
  if ( !_pksb_CompileInnerStatement ( batch, &blab, inloop, false ) )
    goto failure;
  if ( sc->nextsymbol == PKS_SYMB_IDENT &&
       sc->id_space == ID_SPACE_BATCH &&
       sc->nextinteger == pksbatch_ELSE ) {
    jmp.token = PKS_SYMB_BATCH_CMD;
    jmp.key = PKS_BATCHCMD_JUMP;
    jmp.dest = -1;
    jptr = batch->codelgt;
    if ( !_pksb_AddSymbolToCode ( batch, &jmp, sizeof(jumpstatement) ) )
      goto failure;
    sc->GetNextSymbol ( sc );  /* skip else keyword */
    ifst.elseptr = batch->codelgt;
    if ( !_pksb_CompileInnerStatement ( batch, &blab, inloop, false ) )
      goto failure;
    _jmp = (jumpstatement*)&batch->code[jptr+sizeof(unsigned int)];
    _jmp->dest = batch->codelgt;
  }
  else {
    ifst.elseptr = batch->codelgt;
  }
  _ifst = (ifstatement*)&batch->code[ptr+sizeof(unsigned int)];
  _ifst->ifptr = ifst.ifptr;
  _ifst->elseptr = ifst.elseptr;
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_MergeLabelLists ( &_blab->breaks, blab.breaks );
  _pksb_MergeLabelLists ( &_blab->continues, blab.continues );
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksb_CompileIfStatement*/

static boolean _pksb_CompileWhileLoop ( pks_batchexec *batch,
                                        block_labels *_blab )
{
  pks_scanner   *sc;
  block_labels  blab;
  whileloop     whilest, *_whilest;
  jumpstatement jumpst;
  int           ptr;

  sc = &batch->fsc;
  memset ( &blab, 0, sizeof(block_labels) );
  whilest.token = PKS_SYMB_BATCH_CMD;
  whilest.key = PKS_BATCHCMD_WHILE;
  whilest.loopptr = whilest.endptr = 0;
  ptr = batch->codelgt;
  if ( !_pksb_AddSymbolToCode ( batch, &whilest, sizeof(whileloop) ) )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip while keyword */
  if ( !_pksb_CompileParentExpr ( batch ) )
    goto failure;
  whilest.loopptr = batch->codelgt;
  if ( !_pksb_CompileInnerStatement ( batch, &blab, true, false ) )
    goto failure;
  jumpst.token = PKS_SYMB_BATCH_CMD;
  jumpst.key = PKS_BATCHCMD_CONTINUE;
  jumpst.dest = ptr;
  if ( !_pksb_AddSymbolToCode ( batch, &jumpst, sizeof(jumpstatement) ) )
    goto failure;
  whilest.endptr = batch->codelgt;
  _whilest = (whileloop*)&batch->code[ptr+sizeof(unsigned int)];
  _whilest->loopptr = whilest.loopptr;
  _whilest->endptr = whilest.endptr;
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_ResolveBreaks ( batch, whilest.endptr, blab.breaks );
  _pksb_ResolveBreaks ( batch, ptr, blab.continues );
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksb_CompileWhileLoop*/

static boolean _pksb_CompileDoWhileLoop ( pks_batchexec *batch,
                                          block_labels *_blab )
{
  pks_scanner   *sc;
  block_labels  blab;
  whileloop     whilest, *_whilest;
  int           ptr, ptr1;

  sc = &batch->fsc;
  ptr = batch->codelgt;
  sc->GetNextSymbol ( sc );  /* skip do keyword */
  memset ( &blab, 0, sizeof(block_labels) );
  if ( !_pksb_CompileInnerStatement ( batch, &blab, true, false ) )
    goto failure;
  if ( sc->nextsymbol != PKS_SYMB_IDENT ||
       sc->id_space != ID_SPACE_BATCH || sc->nextinteger != pksbatch_WHILE )
    goto failure;
  ptr1 = batch->codelgt;
  whilest.token = PKS_SYMB_BATCH_CMD;
  whilest.key = PKS_BATCHCMD_WHILE;
  whilest.loopptr = ptr;
  whilest.endptr = 0;
  sc->GetNextSymbol ( sc );  /* skip while keyword */
  if ( !_pksb_AddSymbolToCode ( batch, &whilest, sizeof(whileloop) ) )
    goto failure;
  if ( !_pksb_CompileParentExpr ( batch ) )
    goto failure;
  if ( sc->nextsymbol != PKS_SYMB_SEMICOLON )
    goto failure;
  sc->GetNextSymbol ( sc );
  _whilest = (whileloop*)&batch->code[ptr1+sizeof(unsigned int)];
  _whilest->endptr = ptr = batch->codelgt;
  LAST_SEMICOLON = true;
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_ResolveBreaks ( batch, ptr, blab.breaks );
  _pksb_ResolveBreaks ( batch, ptr1, blab.continues );
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksb_CompileDoWhileLoop*/

static boolean _pksb_CompileForLoop ( pks_batchexec *batch,
                                      block_labels *_blab )
{
  pks_scanner   *sc;
  block_labels  blab;
  forloop       forst, *_forst;
  jumpstatement jumpst;
  char          varname[pksbatch_MAXNLGT+1];
  int           ptr;
  short int     intn, op, stop[3];
  int           zero, one;
  auxcomp       *aux;
  forcntname    cntname;

  aux = (auxcomp*)batch->aux;
  sc = &batch->fsc;
  memset ( &blab, 0, sizeof(block_labels) );
  forst.token = PKS_SYMB_BATCH_CMD;
  forst.key = PKS_BATCHCMD_FOR;
  forst.level = aux->fornesting;
  ptr = batch->codelgt;
  if ( !_pksb_AddSymbolToCode ( batch, &forst, sizeof(forloop) ) )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip for keyword */
  if ( sc->nextsymbol != PKS_SYMB_LPAREN )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip opening parenthesis */
  intn = PKS_SYMB_INTEGER;
  one = 1;
  op = PKS_SYMB_NONE;
  if ( sc->nextsymbol == PKS_SYMB_IDENT ) {
    pkv_CopyString ( varname, sc->nextname, pksbatch_MAXNLGT );
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_COLON )
      goto failure;
    sc->GetNextSymbol ( sc );
        /* expression for the initial value */
    forst.initial = batch->codelgt;
    stop[0] = PKS_SYMB_COMMA;
    stop[1] = PKS_SYMB_NONE;
    if ( !_pksb_CompileExpression ( batch, stop ) )
      goto failure;
    if ( sc->nextsymbol != PKS_SYMB_COMMA )
      goto failure;
    sc->GetNextSymbol ( sc );
        /* final value */
    forst.final = batch->codelgt;
    stop[1] = PKS_SYMB_RPAREN;
    stop[2] = PKS_SYMB_NONE;
    if ( !_pksb_CompileExpression ( batch, stop ) )
      goto failure;
    forst.incr = batch->codelgt;
    if ( sc->nextsymbol == PKS_SYMB_COMMA ) {
          /* increment explicitly given */
      sc->GetNextSymbol ( sc );
      if ( !_pksb_CompileExpression ( batch, &stop[1] ) )
        goto failure;
    }
    else {
          /* default increment, 0 to be replaced by -1 or +1 on the fly */
      forst.incr = batch->codelgt;
      if ( !_pksb_AddSymbolToCode ( batch, &intn, sizeof(short int) ) )
        goto failure;
      zero = 0;
      if ( !_pksb_AddToCode ( batch, &zero, sizeof(int) ) )
        goto failure;
      if ( !_pksb_AddSymbolToCode ( batch, &op, sizeof(short int) ) )
        goto failure;
    }
  }
  else {  /* no named variable - just the number of loop executions */
          /* default initial value and increment is 1 */
    varname[0] = 0;
    forst.initial = batch->codelgt;
    if ( !_pksb_AddSymbolToCode ( batch, &intn, sizeof(short int) ) )
      goto failure;
    if ( !_pksb_AddToCode ( batch, &one, sizeof(int) ) )
      goto failure;
    if ( !_pksb_AddSymbolToCode ( batch, &op, sizeof(short int) ) )
      goto failure;
    forst.final = batch->codelgt;
    stop[0] = PKS_SYMB_RPAREN;
    stop[1] = PKS_SYMB_NONE;
    if ( !_pksb_CompileExpression ( batch, stop ) )
      goto failure;
    forst.incr = batch->codelgt;
    if ( !_pksb_AddSymbolToCode ( batch, &intn, sizeof(short int) ) )
      goto failure;
    if ( !_pksb_AddToCode ( batch, &one, sizeof(int) ) )
      goto failure;
    if ( !_pksb_AddSymbolToCode ( batch, &op, sizeof(short int) ) )
      goto failure;
  }
  if ( sc->nextsymbol != PKS_SYMB_RPAREN )
    goto failure;
  sc->GetNextSymbol ( sc );
  forst.loopptr = batch->codelgt;
        /* before compiling the inner statement, add the loop counter name */
        /* to the list, then remove it */
  cntname.name = varname;
  cntname.next = aux->firstcnt;
  aux->firstcnt = &cntname;
  if ( ++aux->fornesting > aux->maxfornesting )
    aux->maxfornesting = aux->fornesting;
  if ( !_pksb_CompileInnerStatement ( batch, &blab, true, false ) )
    goto failure;
  aux->fornesting --;
  aux->firstcnt = cntname.next;
        /* close the inner statement code */
  jumpst.token = PKS_SYMB_BATCH_CMD;
  jumpst.key = PKS_BATCHCMD_CONTINUE;
  jumpst.dest = ptr;
  if ( !_pksb_AddSymbolToCode ( batch, &jumpst, sizeof(jumpstatement) ) )
    goto failure;
  forst.endptr = batch->codelgt;
  _forst = (forloop*)&batch->code[ptr+sizeof(unsigned int)];
  *_forst = forst;
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_ResolveBreaks ( batch, forst.endptr, blab.breaks );
  _pksb_ResolveBreaks ( batch, ptr, blab.continues );
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksb_CompileForLoop*/

static boolean _pksb_CompileSwitchStatement ( pks_batchexec *batch,
                                              block_labels *_blab, boolean inloop )
{
  pks_scanner     *sc;
  block_labels    blab;
  labeldesc       *lbl;
  switchstatement switchst, *_switchst;
  switchcases     swcases, *_swcases;
  jumpstatement   jmpst, *_jmpst;
  int             ptr, ptr1, ptr2, ptr3;
  int             ncases, i, *casetab;

  sc = &batch->fsc;
  casetab = NULL;
  memset ( &blab, 0, sizeof(block_labels) );
  switchst.token = PKS_SYMB_BATCH_CMD;
  switchst.key = PKS_BATCHCMD_SWITCH;
  switchst.cases = -1;
  ptr = batch->codelgt;
  if ( !_pksb_AddSymbolToCode ( batch, &switchst, sizeof(switchstatement) ) )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip switch keyword */
  if ( !_pksb_CompileParentExpr ( batch ) )  
    goto failure;
  blab.defaultloc = -1;
  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  if ( !_pksb_CompileInnerStatement ( batch, &blab, inloop, true ) )
    goto failure;
  ptr1 = batch->codelgt;
  jmpst.token = PKS_SYMB_BATCH_CMD;
  jmpst.key = PKS_BATCHCMD_JUMP;
  jmpst.dest = -1;
  if ( !_pksb_AddSymbolToCode ( batch, &jmpst, sizeof(jumpstatement) ) )
    goto failure;
  ptr2 = batch->codelgt;
  swcases.token = PKS_SYMB_BATCH_CMD;
  swcases.key = PKS_BATCHCMD_CASES;
  swcases.ncases = 0;
  swcases.defaultloc = -1;
  if ( !_pksb_AddSymbolToCode ( batch, &swcases, sizeof(swcases) ) )
    goto failure;
        /* count the cases */
  for ( lbl = blab.cases, ncases = 0; lbl; lbl = lbl->next )
    ncases ++;
  if ( ncases ) {
    ptr3 = batch->codelgt;
    if ( !_pksb_AddToCode ( batch, NULL, 2*ncases*sizeof(int) ) )
      goto failure;
    casetab = (int*)&batch->code[ptr3];
    for ( lbl = blab.cases, i = 0; lbl; lbl = lbl->next ) {
      casetab[i++] = lbl->caseval;
      casetab[i++] = lbl->loc;
    }
    _pksb_DeleteLabelList ( blab.cases );
    if ( ncases > 1 ) {
      if ( pkv_SortFast ( sizeof(int), ID_SIGNED_INT, 2*sizeof(int),
                          0, ncases, casetab ) != SORT_OK )
        goto failure;
    }
  }
  if ( blab.defaultloc < 0 )
    blab.defaultloc = batch->codelgt;
        /* fill the data now available */
  _switchst = (switchstatement*)&batch->code[ptr+sizeof(unsigned int)];
  _switchst->cases = ptr2;
  _jmpst = (jumpstatement*)&batch->code[ptr1+sizeof(unsigned int)];
  _jmpst->dest = batch->codelgt;
  _swcases = (switchcases*)&batch->code[ptr2+sizeof(unsigned int)];
  _swcases->defaultloc = blab.defaultloc;
  _swcases->ncases = ncases;
  _pksb_ResolveBreaks ( batch, batch->codelgt, blab.breaks );
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_MergeLabelLists ( &_blab->continues, blab.continues );
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  if ( casetab ) free ( casetab );
  return false;
} /*_pksb_CompileSwitchStatement*/

static boolean _pksb_CompileCaseLabel ( pks_batchexec *batch,
                                        block_labels *blab )
{
  pks_scanner *sc;
  labeldesc   *lbl;
  pks_calcval x;

  sc = &batch->fsc;
  if ( sc->nextinteger == pksbatch_CASE ) {
    if ( !(lbl = _pksb_NewLabelDesc ( &blab->cases, NULL )) )
      goto failure;
    lbl->loc = batch->codelgt;
    sc->GetNextSymbol ( sc );  /* skip case keyword */
    if ( !pkscalc_LogExpression ( sc, &x ) )
      goto failure;
    if ( x.valtype != PKS_SYMB_INTEGER )
      goto failure;
    lbl->caseval = x.intval;
  }
  else {  /* default */
    if ( blab->defaultloc >= 0 )
      goto failure;
    blab->defaultloc = batch->codelgt;
    sc->GetNextSymbol ( sc );  /* skip default keyword */
  }
  if ( sc->nextsymbol != PKS_SYMB_COLON )
    goto failure;
  sc->GetNextSymbol ( sc );
  LAST_SEMICOLON = false;
  return true;

failure:
  return false;
} /*_pksb_CompileCaseLabel*/

static boolean _pksb_CompileBreakContinueStatement ( pks_batchexec *batch,
                                                     block_labels *blab )
{
  pks_scanner   *sc;
  labeldesc     *lbl;
  jumpstatement jmp;
  int           ptr;

  sc = &batch->fsc;
  jmp.token = PKS_SYMB_BATCH_CMD;
  if ( sc->nextinteger == pksbatch_BREAK ) {
    if ( !(lbl = _pksb_NewLabelDesc ( &blab->breaks, NULL ) ) )
      goto failure;
    jmp.key = PKS_BATCHCMD_JUMP;
  }
  else {
    if ( !(lbl = _pksb_NewLabelDesc ( &blab->continues, NULL ) ) )
      goto failure;
    jmp.key = PKS_BATCHCMD_CONTINUE;
  }
  jmp.dest = -1;
  ptr = batch->codelgt;
  _pksb_AddSymbolToCode ( batch, &jmp, sizeof(jumpstatement) );
  lbl->loc = ptr + sizeof(unsigned int) + (char*)&jmp.dest-(char*)&jmp;
  sc->GetNextSymbol ( sc );  /* skip the keyword */
  if ( sc->nextsymbol != PKS_SYMB_SEMICOLON )
    goto failure;
  sc->GetNextSymbol ( sc );  /* skip the semicolon */
  LAST_SEMICOLON = true;
  return true;

failure:
  return false;
} /*_pksb_CompileBreakContinueStatement*/

static boolean _pksb_CompileGotoStatement ( pks_batchexec *batch,
                                            block_labels *blab )
{
  pks_scanner   *sc;
  labeldesc     *lbl;
  jumpstatement jmp;
  int           ptr;

  sc = &batch->fsc;
  sc->GetNextSymbol ( sc );  /* skip goto keyword */
  if ( sc->nextsymbol != PKS_SYMB_IDENT || sc->id_space == ID_SPACE_BATCH )
    goto failure;
  if ( !(lbl = _pksb_NewLabelDesc ( &blab->gotos, sc->nextname )) )
    goto failure;
  jmp.token = PKS_SYMB_BATCH_CMD;
  jmp.key = PKS_BATCHCMD_GOTO;
  jmp.dest = -1;
  ptr = batch->codelgt;
  if ( !_pksb_AddSymbolToCode ( batch, &jmp, sizeof(jumpstatement) ) )
    goto failure;
  lbl->loc = ptr + sizeof(unsigned int) + (char*)&jmp.dest-(char*)&jmp;
  sc->GetNextSymbol ( sc );  /* skip label name */
  if ( sc->nextsymbol != PKS_SYMB_SEMICOLON )
    goto failure;
  sc->GetNextSymbol ( sc );
  LAST_SEMICOLON = true;
  return true;

failure:
  return false;
} /*_pksb_CompileGotoStatement*/

static boolean _pksb_CompileLabel ( pks_batchexec *batch,
                                    block_labels *blab )
{
  pks_scanner *sc;
  labeldesc   *lbl;

  sc = &batch->fsc;
  sc->GetNextSymbol ( sc );  /* skip label keyword */
  if ( sc->nextsymbol != PKS_SYMB_IDENT )
    goto failure;
  if ( !(lbl = _pksb_NewLabelDesc ( &blab->labels, sc->nextname )) )
    goto failure;
  lbl->loc = batch->codelgt;
  sc->GetNextSymbol ( sc );  /* skip label name */
  if ( sc->nextsymbol != PKS_SYMB_COLON )
    goto failure;
  sc->GetNextSymbol ( sc );
  LAST_SEMICOLON = false;
  return true;

failure:
  return false;
} /*_pksb_CompileLabel*/

static boolean _pksb_CompileKeyStatement ( pks_batchexec *batch,
                                           block_labels *blab,
                                           boolean inloop, boolean inswitch )
{
  pks_scanner *sc;

  sc = &batch->fsc;
  batch->codelgt -= sizeof(unsigned int) +
                    sizeof(short int);  /* withdraw the last token */
  switch ( sc->nextinteger ) {
case pksbatch_CASE:
case pksbatch_DEFAULT:
    if ( !inswitch )
      goto failure;
    if ( !_pksb_CompileCaseLabel ( batch, blab ) )
      goto failure;
    break;

case pksbatch_CONTINUE:
    if ( !inloop )
      goto failure;
case pksbatch_BREAK:
    if ( !inloop && !inswitch )
      goto failure;
    if ( !_pksb_CompileBreakContinueStatement ( batch, blab ) )
      goto failure;
    break;

case pksbatch_DO:
    if ( !_pksb_CompileDoWhileLoop ( batch, blab ) )
      goto failure;
    break;

case pksbatch_FOR:
    if ( !_pksb_CompileForLoop ( batch, blab ) )
      goto failure;
    break;

case pksbatch_GOTO:
    if ( !_pksb_CompileGotoStatement ( batch, blab ) )
      goto failure;
    LAST_SEMICOLON = true;
    break;

case pksbatch_IF:
    if ( !_pksb_CompileIfStatement ( batch, blab, inloop ) )
      goto failure;
    break;

case pksbatch_LABEL:
    if ( !_pksb_CompileLabel ( batch, blab ) )
      return false;
    break;

case pksbatch_SWITCH:
    if ( !_pksb_CompileSwitchStatement ( batch, blab, inloop ) )
      goto failure;
    break;

case pksbatch_WHILE:
    if ( !_pksb_CompileWhileLoop ( batch, blab ) )
      goto failure;
    break;

default:
    goto failure;
  }
  return true;

failure:
  return false;
} /*_pksb_CompileKeyStatement*/

static void _pksb_SetupBannedTokens ( short int *outb, short int stop,
                                      short int *banned )
{
  int i;

        /* this is to forbid keywords in expressions */
  if ( outb[0] == PKS_SYMB_BATCH_CMD )
    { banned[0] = PKS_SYMB_BATCH_CMD; i = 1; }
  else
    i = 0;
        /* this is to forbid wrong closing parentheses */
  switch ( stop ) {
case PKS_SYMB_RPAREN:
    banned[i++] = PKS_SYMB_RBRACKET;
    banned[i++] = PKS_SYMB_RBRACE;
    break;
case PKS_SYMB_RBRACKET:
    banned[i++] = PKS_SYMB_RPAREN;
    banned[i++] = PKS_SYMB_RBRACE;
    break;
case PKS_SYMB_RBRACE:
    banned[i++] = PKS_SYMB_RPAREN;
    banned[i++] = PKS_SYMB_RBRACKET;
    break;
default:
    banned[i++] = PKS_SYMB_RBRACKET;
    banned[i++] = PKS_SYMB_RBRACE;
    banned[i++] = PKS_SYMB_RBRACKET;
    break;
  }
  banned[i] = PKS_SYMB_NONE;
} /*_pksb_SetupBannedTokens*/

static boolean _pksb_CompileTokenSequence ( pks_batchexec *batch,
                                            block_labels *_blab,
                                            short int *stop, short int *banned,
                                            boolean inloop, boolean inswitch )
{
  pks_scanner  *sc;
  block_labels blab;
  boolean      forc;
  short int    newbanned[5], newstop[2];
  int          i;

  sc = &batch->fsc;
  memset ( &blab, 0, sizeof(block_labels) );
  if ( inswitch )
    blab.defaultloc = -1;
  for (;;) {
    LAST_SEMICOLON = false;
    if ( !_pksb_AddSymbolToCode ( batch, &sc->nextsymbol, sizeof(short int) ) )
      goto failure;
    switch ( sc->nextsymbol ) {
  case PKS_SYMB_ERROR:
      goto failure;
  case PKS_SYMB_LPAREN:
  case PKS_SYMB_LBRACKET:
  case PKS_SYMB_LBRACE:
      newstop[0] = sc->nextsymbol + 1;
      newstop[1] = PKS_SYMB_NONE;
      sc->GetNextSymbol ( sc );
      _pksb_SetupBannedTokens ( banned, newstop[0], newbanned );
      if ( !_pksb_CompileTokenSequence ( batch, &blab, newstop, newbanned,
                                         inloop, false ) )
        goto failure;
      if ( sc->nextsymbol != newstop[0] )
        goto failure;
      if ( !_pksb_AddSymbolToCode ( batch, &sc->nextsymbol, sizeof(short int) ) )
        goto failure;
      break;
  case PKS_SYMB_IDENT:
      switch ( sc->id_space ) {
    case ID_SPACE_BATCH:
        if ( !_pksb_CompileKeyStatement ( batch, &blab, inloop, inswitch ) )
          goto failure;
        else {
          if ( LAST_SEMICOLON && stop[0] == PKS_SYMB_SEMICOLON )
            goto success;
          goto check_stop;  /* do not skip the token, it's already done */
        }
    default:
        if ( sc->nextsymbol == PKS_SYMB_IDENT ) {
          if ( !_pksb_EnterForCounter ( batch, &forc ) )
            goto failure;
          if ( forc )
            break;
        }
        if ( !_pksb_AddLiteralToCode ( batch ) )
          goto failure;
        break;
      }
      break;
  case PKS_SYMB_INTEGER:
  case PKS_SYMB_CHAR:
  case PKS_SYMB_FLOAT:
  case PKS_SYMB_STRING:
  case PKS_SYMB_OTHER:
      if ( !_pksb_AddLiteralToCode ( batch ) )
        goto failure;
      break;
  case PKS_SYMB_SEMICOLON:
      LAST_SEMICOLON = true;
      break;
  case PKS_SYMB_EOF:
      goto success;
  default:
      break;
    }
    sc->GetNextSymbol ( sc );
check_stop:
    for ( i = 0; stop[i] != PKS_SYMB_NONE; i++ )
      if ( sc->nextsymbol == stop[i] )
        goto success;
    for ( i = 0; banned[i] != PKS_SYMB_NONE; i++ )
      if ( sc->nextsymbol == banned[i] )
        goto failure;
  }

success:
  if ( sc->nextsymbol == PKS_SYMB_SEMICOLON ) {
         /* the ending semicolon has to be stored in the code */
    if ( !_pksb_AddSymbolToCode ( batch, &sc->nextsymbol, sizeof(short int) ) )
      goto failure;
    sc->GetNextSymbol ( sc );
    LAST_SEMICOLON = true;
  }
  _pksb_ResolveGotos ( batch, &blab.labels, &blab.gotos );
  _pksb_MergeLabelLists ( &_blab->gotos, blab.gotos );
  _pksb_MergeLabelLists ( &_blab->breaks, blab.breaks );
  _pksb_MergeLabelLists ( &_blab->continues, blab.continues );
  _blab->cases = blab.cases;
  _blab->defaultloc = blab.defaultloc;
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksb_CompileTokenSequence*/

static boolean _pksbatch_IncludeFileProc ( pks_scanner *sc, boolean open )
{
  pks_batchexec    *batch;
  pks_file         *f1, *f2;
  include_filename ifn;
  include_end      eof;
  int              place;

  if ( !(f1 = sc->infile) )
    return false;
  f2 = f1->inclink;
  batch = sc->libdata;
  if ( open ) {  /* source file has been opened */
    place = f2 ? PKS_SCANNER_MAX_LINE_LENGTH*f2->linenum + f2->colnum :
                 UINT_MAX;
    f1->ip = batch->codelgt;
    ifn.token = PKS_SYMB_BATCH_CMD;
    ifn.key = PKS_BATCHCMD_INCLUDE;
    if ( !_pksb_AddToCode ( batch, &place, sizeof(unsigned int) ) )
      return false;
    if ( !_pksb_AddToCode ( batch, &ifn, sizeof(include_filename) ) )
      return false;
    return _pksb_AddToCode ( batch, f1->filename, strlen(f1->filename)+1 );
  }
  else if ( f2 ) {  /* source file is being closed */
    place = PKS_SCANNER_MAX_LINE_LENGTH*f2->linenum + f2->colnum;
    eof.token = PKS_SYMB_BATCH_CMD;
    eof.key = PKS_BATCHCMD_ENDINCL;
    eof.namepos = f2->ip;
    if ( !_pksb_AddToCode ( batch, &place, sizeof(unsigned int) ) )
      return false;
    return _pksb_AddToCode ( batch, &eof, sizeof(eof) );
  }
  else
    return false;
} /*_pksbatch_IncludeFileProc*/

static boolean _pksbatch_Compile ( pks_batchexec *batch )
{
  auxcomp      aux;
  block_labels blab;
  int          ptr;
  short int    stop[2], banned, eof;
  char         *auxc;

  memset ( &aux, 0, sizeof(aux) );
  batch->aux = (void*)&aux;
  batch->fsc.GetNextSymbol ( &batch->fsc );
  memset ( &blab, 0, sizeof(block_labels) );
  stop[0] = PKS_SYMB_EOF;
  stop[1] = banned = PKS_SYMB_NONE;
  if ( !_pksb_CompileTokenSequence ( batch, &blab, stop, &banned,
                                     false, false ) )
    goto failure;
  if ( blab.gotos || blab.labels ||
       blab.breaks || blab.continues || blab.cases )
    goto failure;
  eof = PKS_SYMB_EOF;
  if ( !_pksb_AddSymbolToCode ( batch, &eof, sizeof(short int) ) )
    goto failure;
  ptr = batch->codelgt;
  if ( aux.maxfornesting > 0 ) {  /* allocate storage space for for loop counters */
    if ( !_pksb_AddToCode ( batch, NULL, 3*aux.maxfornesting*sizeof(int) ) )
      goto failure;
  }
  if ( batch->codelgt < batch->buflgt ) {
    if ( (auxc = realloc ( batch->code, batch->codelgt )) ) {
      batch->code = auxc;
      batch->buflgt = batch->codelgt;
    }
  }
  batch->codelgt = ptr;
  batch->var = (int*)&batch->code[ptr];
  batch->aux = NULL;
  batch->ip = 0;
  return true;

failure:
  _pksb_DeleteLabelLists ( &blab );
  return false;
} /*_pksbatch_Compile*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean pksbatch_InitExecution ( pks_batchexec *batch,
                        unsigned int options,
                        const char *filename,
                        int inbuflength, int maxnamelength,
                        const char *bcomment, const char *ecomment,
                        char *params,
                        int (*InputData)(pks_scanner*),
                        boolean (*AppIdentFind)(pks_scanner*),
                        boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                        void *userdata )
{
  memset ( batch, 0, sizeof(pks_batchexec) );
  if ( !pks_InitPreprocessor ( &batch->fsc, options, inbuflength, maxnamelength,
                               bcomment, ecomment, InputData,
                               pks_GetNextPSymbol, _pksb_AppIdentFind,
                               _pksb_IdentHook, userdata ) )
    return false;
  batch->code = NULL;
  batch->buflgt = batch->codelgt = 0;
  batch->fsc.IncludeFileHook = _pksbatch_IncludeFileProc;
  batch->fsc.libdata = batch;
  if ( !AppIdentFind ) AppIdentFind = pks_AppIdentFind;
  batch->AppIdentFind = AppIdentFind;
  if ( !AppIdentCalc ) AppIdentCalc = pks_AppIdentCalc;
  batch->AppIdentCalc = AppIdentCalc;
  if ( !pks_OpenInputFile ( &batch->fsc, filename ) )
    goto failure;
  if ( !_pksbatch_ProcessParameters ( batch, options, params, maxnamelength ) )
    goto failure;
  pkv_Tic ( &batch->ticks );
  if ( !_pksbatch_Compile ( batch ) )
    goto failure;
  pks_ShutDownPreprocessor ( &batch->fsc );
  memset ( &batch->csc, 0, sizeof(pks_scanner) );
  batch->fsc.nextname = malloc ( maxnamelength+1 );
  batch->csc.nextname = malloc ( maxnamelength+1 );
  if ( !batch->fsc.nextname || !batch->csc.nextname )
    goto failure;
  batch->csc.libdata = batch->fsc.libdata = batch;
  batch->fsc.GetNextSymbol = _pksbatch_GetNextSymbolFSC;
  batch->csc.GetNextSymbol = _pksbatch_GetNextSymbolCSC;
  batch->csc.AppIdentFind = batch->fsc.AppIdentFind = AppIdentFind;
  batch->csc.AppIdentCalc = batch->fsc.AppIdentCalc = _pksb_IdentHook;
  batch->ip = 0;
  batch->fetch_next = true;
  batch->csc.GetNextSymbol ( &batch->csc );
  return true;

failure:
  pks_ShutDownPreprocessor ( &batch->fsc );
  pks_ShutDownPreprocessor ( &batch->csc );
  memset ( batch, 0, sizeof(pks_batchexec) );
  return false;
} /*pksbatch_InitExecution*/

