
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <stdio.h>

#include "pkvaria.h"
#include "avltree.h"

/* ///////////////////////////////////////////////////////////////////////// */
void AVLInit ( AVLtree *tree, boolean no_double,
               char (*compkeys)(AVLtnode*,void*),
               AVLtnode *(*allocnode)(void*),
               void (*freenode)(AVLtnode*) )
{
  tree->root = tree->last = NULL;
  tree->no_double = no_double;
  tree->error_code = AVLERROR_NONE;
  tree->compkeys = compkeys;
  tree->allocnode = allocnode;
  tree->freenode = freenode;
} /*AVLInit*/

static void _AVLRotRight ( AVLtnode **root )
{
  AVLtnode *a, *b;
  char     a0, b0, b1;

  b = *root;  a = b->left;
  b->left = a->right;  a->right = b;
  *root = a;
  a0 = a->eqi;  b0 = b->eqi;
  if ( a0 >= 0 ) {
    b1 = b0 - a0 -1;
    a->eqi = b1 >= 0 ? a0 - 1 : b0 - 2;
  }
  else {
    b1 = b0 - 1;
    a->eqi = b1 >= 0 ? a0 - 1 : a0 + b0 - 2;
  }
  b->eqi = b1;
} /*_AVLRotRight*/

static void _AVLRotLeft ( AVLtnode **root )
{
  AVLtnode *a, *b;
  char     a0, a1, b0;

  a = *root;  b = a->right;
  a->right = b->left;  b->left = a;
  *root = b;
  a0 = a->eqi;  b0 = b->eqi;
  if ( b0 <= 0 ) {
    a1 = a0 - b0 + 1;
    b->eqi = a1 <= 0 ? b0 + 1 : a0 + 2;
  }
  else {
    a1 = a0 + 1;
    b->eqi = a1 <= 0 ? b0 + 1 : a0 + b0 + 2;
  }
  a->eqi = a1;
} /*_AVLRotLeft*/

static boolean _AVLInsert ( AVLtree *tree, AVLtnode **node, void *data )
{
  char c;

  if ( !(*node) ) {
    *node = tree->last = tree->allocnode ( data );
    if ( !(*node) ) {
      tree->error_code = AVLERROR_ALLOC;
      return false;
    }
    (*node)->left = (*node)->right = NULL;
    (*node)->eqi = 0;
    return true;
  }
  else {
    c = tree->compkeys ( *node, data );
    if ( c < -1 || c > +1 ) {
      tree->error_code = AVLERROR_COMPARE;
      return false;
    }
    if ( c == 0 && tree->no_double ) {
      tree->error_code = AVLERROR_DOUBLE_KEY;
      return false;
    }
    if ( c >= 0 ) {
      if ( _AVLInsert ( tree, &(*node)->left, data ) ) {
        (*node)->eqi ++;
        if ( (*node)->eqi == 2 ) {
          if ( (*node)->left->eqi < 0 )
            _AVLRotLeft ( &(*node)->left );
          _AVLRotRight ( node );
          return false;
        }
        else return (*node)->eqi == 1;
      }
      else return false;
    }
    else {
      if ( _AVLInsert ( tree, &(*node)->right, data ) ) {
        (*node)->eqi --;
        if ( (*node)->eqi == -2 ) {
          if ( (*node)->right->eqi > 0 )
            _AVLRotRight ( &(*node)->right );
          _AVLRotLeft ( node );
          return false;
        }
        else return (*node)->eqi == -1;
      }
      else return false;
    }
  }
} /*_AVLInsert*/

boolean AVLInsert ( AVLtree *tree, void *data )
{
  return _AVLInsert ( tree, &tree->root, data );
} /*AVLInsert*/

static boolean _AVLDelete ( AVLtree *tree, AVLtnode **node,
                            AVLtnode **todelete, AVLtnode ***swl,
                            void *data, int level )
{
  char     c;
  AVLtnode *tn, *a;

  if ( !(*node) ) {
    tree->error_code = AVLERROR_DELETE;  /* node to delete not found */
    return false;
  }
  tn = *node;
  if ( todelete ) {
    if ( !tn->right ) {  /* swap the nodes and delete */
      a = *todelete;
      *node = tn->left;
      *todelete = tn;
      tn->left = a->left;  tn->right = a->right;
      tn->eqi = a->eqi;
      if ( a == tree->last )
        tree->last = NULL;
      tree->freenode ( a );
      *swl = &tn->left;
      return true;
    }
    else if ( _AVLDelete ( tree, &tn->right, todelete, swl, data, level+1 ) )
      goto the_right;
    else
      return false;
  }
  else {
    c = tree->compkeys ( tn, data );
    if ( c < -1 || c > +1 ) {
      tree->error_code = AVLERROR_COMPARE;
      return false;
    }
    if ( c == 0 ) {
      todelete = node;
      if ( !tn->right )
        *node = tn->left;
      else if ( !tn->left )
        *node = tn->right;
      else
        goto tough_case;
      tree->freenode ( tn );
      return true;
    }
tough_case:
    if ( c >= 0 ) {
      if ( _AVLDelete ( tree, &tn->left, todelete, swl, data, level+1 ) ) {
        tn = *node;
        tn->eqi --;
        if ( tn->eqi == -2 ) {
          if ( tn->right->eqi > 0 )
            _AVLRotRight ( &tn->right );
          _AVLRotLeft ( node );
        }
        return (*node)->eqi == 0;
      }
      else return false;
    }
    else {
      if ( _AVLDelete ( tree, &tn->right, todelete, swl, data, level+1 ) ) {
the_right:
        if ( *swl ) {
          if ( **swl == tn ) { node = *swl;  tn = *node; }
        }
        tn->eqi ++;
        if ( tn->eqi == +2 ) {
          if ( tn->left->eqi < 0 )
            _AVLRotLeft ( &tn->left );
          _AVLRotRight ( node );
        }
        return (*node)->eqi == 0;
      }
      else return false;
    }
  }
} /*_AVLDelete*/

boolean AVLDelete ( AVLtree *tree, void *data )
{
  AVLtnode **swl;

  swl = NULL;
  return _AVLDelete ( tree, &tree->root, NULL, &swl, data, 0 );
} /*AVLDelete*/

AVLtnode *AVLFind ( AVLtree *tree, void *data )
{
  AVLtnode *node;

  node = tree->root;
  while ( node ) {
    switch ( tree->compkeys ( node, data) ) {
  case -1:
      node = node->right;
      break;
  case  0:
      return tree->last = node;
  case +1:
      node = node->left;
      break;
  default:
      tree->error_code = AVLERROR_COMPARE;
      return NULL;
    }
  }
  return NULL;
} /*AVLFind*/

static void _AVLTraverse ( AVLtree *tree, AVLtnode *node, boolean descending,
                           void *usrdata, void (*process)(void*,AVLtnode*) )
{
  if ( descending ) {
    if ( node->right )
      _AVLTraverse ( tree, node->right, descending, usrdata, process );
    process ( usrdata, node );
    if ( node->left )
      _AVLTraverse ( tree, node->left, descending, usrdata, process );
  }
  else {
    if ( node->left )
      _AVLTraverse ( tree, node->left, descending, usrdata, process );
    process ( usrdata, node );
    if ( node->right )
      _AVLTraverse ( tree, node->right, descending, usrdata, process );
  }
} /*_AVLTraverse*/

void AVLTraverse ( AVLtree *tree, boolean descending, void *usrdata,
                   void (*process)(void*,AVLtnode*) )
{
  if ( tree->root )
    _AVLTraverse ( tree, tree->root, descending, usrdata, process );
} /*AVLTraverse*/

static void _AVLDestroy ( AVLtree *tree, AVLtnode *tnode )
{
  if ( tnode->left ) _AVLDestroy ( tree, tnode->left );
  if ( tnode->right ) _AVLDestroy ( tree, tnode->right );
  tree->freenode ( tnode );
} /*_AVLDestroy*/

void AVLDestroy ( AVLtree *tree )
{
  if ( tree->root ) {
    _AVLDestroy ( tree, tree->root );
    tree->root = tree->last = NULL;
  }
} /*AVLDestroy*/

