
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkgeom.h"
#include "multibs.h"

#include "xgedit.h"
#include "xgeprivate.h"

XImage *xge_CreateXImage ( short w, short h )
{
  int    nplanes;
  XImage *ximage;
  char   *imagedata;

  nplanes = XDisplayPlanes ( xgedisplay, xgescreen );
  ximage = XCreateImage ( xgedisplay, xgevisual, nplanes, ZPixmap, 0,
                          NULL, w, h, 8, 0 );
  if ( !ximage )
    return NULL;
  imagedata = malloc ( (int)h*ximage->bytes_per_line );
  if ( !imagedata ) {
    XDestroyImage ( ximage );
    return NULL;
  }
  else
    ximage->data = imagedata;
  return ximage;
} /*xge_CreateXImage*/

void xge_DestroyXImage ( XImage *ximage )
{
  if ( ximage ) {
    if ( ximage->data ) {
      free ( ximage->data );
      ximage->data = NULL;
    }
    XDestroyImage ( ximage );
  }
} /*xge_DestroyXImage*/

static unsigned int _xge_GetMaskShift ( unsigned int mask, unsigned short *shift )
{
  unsigned int am;
  short        sh;

  am = 0x80000000;
  sh = 24;
  while ( !(am & mask) && sh ) {
    am >>= 1;
    sh --;
  }
  *shift = sh;
  return mask;
} /*_xge_GetMaskShift*/

/* ///////////////////////////////////////////////////////////////////////// */
/* The procedure xge_GetImage reads an image from the pixmap associated with */
/* the current xge window and stores it in the array passed by an applicaton.*/
/* Parameters:                                                               */
/* w, h - width and height of the image to read,                             */
/* x, y - position (upper left corner) in the window,                        */
/* rgb - array of length at least w*h*3 for the byte r, g, b components,     */
/* ximage - pointer to an XWindow image used as a buffer, must be in ZPixmap */
/*          format, with image data of type xgecolour_int and of suitable    */
/*          size; if NULL then the procedure will allocate it and then       */
/*          destroy.                                                         */
/* Originally this procedure read the image from the window, not the pixmap. */
/* Then the application might crash, if the window was not mapped on the     */
/* current screen.                                                           */
boolean xge_GetImage ( short w, short h, short x, short y,
                       unsigned char *rgb,
                       XImage *ximage )
{
  XImage         *image;
  char           *ip;
  unsigned int   *pixel;
  int            bpp, bpl, i;
  unsigned int   rm, gm, bm;  /* masks */
  unsigned short rs, gs, bs;  /* shifts */

  if ( ximage )
    image = ximage;
  else {
    image = xge_CreateXImage ( w, h );
    if ( !image )
      return false;
  }
  XGetSubImage ( xgedisplay, xgepixmap, x, y, w, h,
                 0xFFFFFFFF, ZPixmap, image, 0, 0 );
  bpl = image->bytes_per_line;
  bpp = image->bits_per_pixel/8;
  rm = _xge_GetMaskShift ( (unsigned int)image->red_mask, &rs );
  gm = _xge_GetMaskShift ( (unsigned int)image->green_mask, &gs );
  bm = _xge_GetMaskShift ( (unsigned int)image->blue_mask, &bs );
  for ( y = i = 0;  y < h;  y++ ) {
    ip = &image->data[y*bpl];
    for ( x = 0;  x < w;  x++, i += 3 ) {
      pixel = (unsigned int*)&ip[bpp*x];
      rgb[i]   = (*pixel & rm) >> rs;
      rgb[i+1] = (*pixel & gm) >> gs;
      rgb[i+2] = (*pixel & bm) >> bs;
    }
  }
  if ( !ximage )
    xge_DestroyXImage ( image );
  return true;
} /*xge_GetImage*/

