
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */ 
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkvaria.h"
#include "xgledit.h"
#include "xgeprivate.h"
#include "xgleprivate.h"

extern char *ScratchPtr;

/* ///////////////////////////////////////////////////////////////// */
/* two types of objects need to be drawn often enough to prepare     */
/* the OpenGL mechanism for it in the immediate mode; a line segment */
/* and text; it would certainly be better to use XWindow, but not    */
/* when these are elements of the background for any graphics        */
/* displayed by OpenGL on that background.                           */
/* ///////////////////////////////////////////////////////////////// */
#include "xglegldraw.glsl.c"

static const char *gcstr[5] =
  { "xgleGC", "xgleGC.pm", "xgleGC.fg", "xgleGC.bk", "xgleGC.h" };
static const char *fontstr[6] =
  { "xgleFont", "xgleFont.chw", "xgleFont.chh",
    "xgleFont.chf", "xgleFont.chl", "xgleFont.glyphs" };
static const char *textstr[5] =
  { "xgleText", "xgleText.x", "xgleText.y", "xgleText.l", "xgleText.text" };

static struct {
    GLuint prog0, prog1;
    GLuint sh[4];
    GLuint gcindex[2];
    GLint  gcofs[5];
    GLuint fontindex;
    GLint  fontofs[5];
    GLuint textindex;
    GLint  textofs[4];
    GLint  gcsize, fontsize, textsize;
    GLuint gcbuf, textbuf,
           font12x6buf, font18x10buf;
    GLuint vao, vbo;
    byte   chw, chh;
  } glres;

static boolean _xgle_InitPrivatePrograms ( void )
{
  GLuint ind[5];

        /* compile the shaders and link the programs */
  glres.sh[0] = xgle_CompileShaderStrings ( GL_VERTEX_SHADER, 3, _shader_v );
  glres.sh[1] = xgle_CompileShaderStrings ( GL_FRAGMENT_SHADER, 3, _shader_f0 );
  glres.sh[2] = glres.sh[0];
  glres.sh[3] = xgle_CompileShaderStrings ( GL_FRAGMENT_SHADER, 6, _shader_f1 );
  glres.prog0 = xgle_LinkProgram ( 2, glres.sh );
  glres.prog1 = xgle_LinkProgram ( 2, &glres.sh[2] );
        /* get access to the uniform blocks */
  glres.gcindex[0] = glGetUniformBlockIndex ( glres.prog0, gcstr[0] );
  glres.gcindex[1] = glGetUniformBlockIndex ( glres.prog1, gcstr[0] );
  glGetActiveUniformBlockiv ( glres.prog1, glres.gcindex[1],
                              GL_UNIFORM_BLOCK_DATA_SIZE, &glres.gcsize );
  glGetUniformIndices ( glres.prog1, 4, &gcstr[1], ind );
  glGetActiveUniformsiv ( glres.prog1, 4, ind, GL_UNIFORM_OFFSET, glres.gcofs );

  glres.fontindex = glGetUniformBlockIndex ( glres.prog1, fontstr[0] );
  glGetActiveUniformBlockiv ( glres.prog1, glres.fontindex,
                              GL_UNIFORM_BLOCK_DATA_SIZE, &glres.fontsize );
  glGetUniformIndices ( glres.prog1, 5, &fontstr[1], ind );
  glGetActiveUniformsiv ( glres.prog1, 5, ind, GL_UNIFORM_OFFSET, glres.fontofs );

  glres.textindex = glGetUniformBlockIndex ( glres.prog1, textstr[0] );
  glGetActiveUniformBlockiv ( glres.prog1, glres.textindex,
                               GL_UNIFORM_BLOCK_DATA_SIZE, &glres.textsize );
  glGetUniformIndices ( glres.prog1, 4, &textstr[1], ind );
  glGetActiveUniformsiv ( glres.prog1, 4, ind, GL_UNIFORM_OFFSET, glres.textofs );
        /* allocate and bind the GC and text buffers */
  glUniformBlockBinding ( glres.prog0, glres.gcindex[0], XGLE_BINDPOINT_GC );
  glUniformBlockBinding ( glres.prog1, glres.gcindex[1], XGLE_BINDPOINT_GC );
  glGenBuffers ( 1, &glres.gcbuf );
  glBindBuffer ( GL_UNIFORM_BUFFER, glres.gcbuf );
  glBufferData ( GL_UNIFORM_BUFFER, glres.gcsize, NULL, GL_DYNAMIC_DRAW );
  glBindBufferBase ( GL_UNIFORM_BUFFER, XGLE_BINDPOINT_GC, glres.gcbuf );

  glUniformBlockBinding ( glres.prog1, glres.textindex, XGLE_BINDPOINT_TEXT );
  glGenBuffers ( 1, &glres.textbuf );
  glBindBuffer ( GL_UNIFORM_BUFFER, glres.textbuf );
  glBufferData ( GL_UNIFORM_BUFFER, glres.textsize, NULL, GL_DYNAMIC_DRAW );
  glBindBufferBase ( GL_UNIFORM_BUFFER, XGLE_BINDPOINT_TEXT, glres.textbuf );

  glUniformBlockBinding ( glres.prog1, glres.fontindex, XGLE_BINDPOINT_FONT );
        /* create the vertex array object to draw the lines and line segments */
  glGenVertexArrays ( 1, &glres.vao );
  glBindVertexArray ( glres.vao );
  glGenBuffers ( 1, &glres.vbo );
  glBindBuffer ( GL_ARRAY_BUFFER, glres.vbo );
  glBufferData ( GL_ARRAY_BUFFER, 8*sizeof(GLfloat), NULL, GL_DYNAMIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 2, GL_FLOAT, GL_FALSE, 2*sizeof(GLfloat), (GLvoid*)0 );
  glBindVertexArray ( 0 );

  return true;
} /*_xgle_InitPrivatePrograms*/

/* ///////////////////////////////////////////////////////////////////////// */
boolean xgle_InitPrivateGLObjects ( void )
{
  boolean no_scratch;

  if ( (no_scratch = (ScratchPtr == NULL)) )
    pkv_InitScratchMem ( 1048576 );
  if ( !_xgle_InitPrivatePrograms () )
    goto failure;
  if ( !_xgle_Init12x6font ( glres.fontofs, &glres.font12x6buf ) )
    goto failure;
  if ( !_xgle_Init18x10font ( glres.fontofs, &glres.font18x10buf ) )
    goto failure;
  xgle_Choose12x6font ();
  if ( no_scratch )
    pkv_DestroyScratchMem ();
  return true;

failure:
  if ( no_scratch )
    pkv_DestroyScratchMem ();
  return false;
} /*xgle_InitPrivateGLObjects*/

/* ///////////////////////////////////////////////////////////////////////// */
void xgle_SetupXGLCoordinates ( GLint w, GLint h )
{
  GLfloat pm[16];

  glBindBuffer ( GL_UNIFORM_BUFFER, glres.gcbuf );
  xgle_Orthof ( -0.5, (float)w-0.5, (float)h-0.5, -0.5, -1.0, 1.0, pm, NULL );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.gcofs[0], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.gcofs[3], sizeof(GLint), &h );
} /*xgle_SetupXGLCoordinates*/

void xgle_SetForeground ( unsigned int foreground, byte alpha )
{
  GLfloat fg[4];

  xge_GetPixelColourf ( foreground, &fg[0], &fg[1], &fg[2] );
  fg[3] = (float)alpha/255.0;
  glBindBuffer ( GL_UNIFORM_BUFFER, glres.gcbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.gcofs[1], 4*sizeof(GLfloat), fg );
} /*xgle_SetForeground*/

void xgle_SetBackground ( unsigned int background, byte alpha )
{
  GLfloat fg[4];

  xge_GetPixelColourf ( background, &fg[0], &fg[1], &fg[2] );
  fg[3] = (float)alpha/255.0;
  glBindBuffer ( GL_UNIFORM_BUFFER, glres.gcbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.gcofs[2], 4*sizeof(GLfloat), fg );
} /*xgle_SetBackground*/

void xgle_DrawLine ( GLfloat x0, GLfloat y0, GLfloat x1, GLfloat y1 )
{
  GLfloat v[4];

  v[0] = x0;  v[1] = y0;  v[2] = x1;  v[3] = y1;
  glBindBuffer ( GL_ARRAY_BUFFER, glres.vbo );
  glBufferSubData ( GL_ARRAY_BUFFER, 0, 4*sizeof(GLfloat), v );
  glBindVertexArray ( glres.vao );
  glUseProgram ( glres.prog0 );
  glDrawArrays ( GL_LINES, 0, 2 );
  glUseProgram ( 0 );
} /*xgle_DrawLine*/

void xgle_DrawString ( const char *string, int x, int y )
{
  GLuint lgt;
  GLfloat r[8];

  lgt = strlen ( string );
  if ( !lgt )
    return;
  if ( lgt > XGLE_MAX_STRING_LENGTH )
    lgt = XGLE_MAX_STRING_LENGTH;
  glBindBuffer ( GL_UNIFORM_BUFFER, glres.textbuf );
  y -= glres.chh;
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.textofs[0], sizeof(GLint), &x );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.textofs[1], sizeof(GLint), &y );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.textofs[2], sizeof(GLint), &lgt );
  glBufferSubData ( GL_UNIFORM_BUFFER, glres.textofs[3], lgt, string );
  r[0] = r[6] = (GLfloat)x;                  r[1] = r[3] = (GLfloat)y;
  r[2] = r[4] = (GLfloat)(x+lgt*glres.chw);  r[5] = r[7] = (GLfloat)(y+glres.chh);
  glBindBuffer ( GL_ARRAY_BUFFER, glres.vbo );
  glBufferSubData ( GL_ARRAY_BUFFER, 0, 8*sizeof(GLfloat), r );  
  glUseProgram ( glres.prog1 );
  glBindVertexArray ( glres.vao );
  glDrawArrays ( GL_TRIANGLE_FAN, 0, 4 );
  glUseProgram ( 0 );
} /*xgle_DrawString*/

void xgle_Choose12x6font ( void )
{  
  glBindBufferBase ( GL_UNIFORM_BUFFER, XGLE_BINDPOINT_FONT, glres.font12x6buf );
  glres.chw = 6;  glres.chh = 12;
} /*xgle_Choose12x6font*/

void xgle_Choose18x10font ( void )
{  
  glBindBufferBase ( GL_UNIFORM_BUFFER, XGLE_BINDPOINT_FONT, glres.font18x10buf );
  glres.chw = 10;  glres.chh = 18;
} /*xgle_Choose18x10font*/

